#ifndef _FREQ_H
#define _FREQ_H

extern int mcpGetFreq8363(int note);
extern int mcpGetFreq6848(int note);
extern int mcpGetNote8363(unsigned int frq);
extern int mcpGetNote6848(unsigned int frq);

#endif
