/* OpenCP Module Player
 * copyright (c) '94-'10 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * assembler routines for Quality Mixer
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * revision history: (please note changes here)
 *  -nb980510   Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *    -first release
 *  -kb980717 Tammo Hinrichs <opencp@gmx.net>
 *    -some pentium optimization on inner loops
 *  -kbwhenever Tammo Hinrichs <opencp@gmx.net>
 *    -used CS addressing here and there for more optimal cache usage
 *  -ss04????
 */

#include "config.h"
#include <stdio.h>
#include "types.h"
#include "dwmix.h"
#include "dwmixqa.h"

#if defined(__PIC__) && defined(I386_ASM)
#warning I386_ASM is disabled in non FPU mixers when compiled PIC
#undef I386_ASM
#endif

#ifdef I386_ASM

#include "dwmixqa_x86.c"

#else

#include "dwmixqa_c.c"

#endif
